--***********************************************************************
--***********************************************************************
-- INTRODUCTION
--***********************************************************************
--***********************************************************************

/*

Federated Query Simulation
Copyright (C) 2014 Griffin M Weber

Griffin M Weber, MD, PhD
weber@hms.harvard.edu

October 1, 2014

This script simulates a federated data network with 4000 hospitals.
The parameters for the simulation are based on the actual Shared
Health Research Information Network (SHRINE) that was developed at
Harvard Medical School and launched in 2008. The simulation models
the network running for 1000 days, and it repeats the experiments
1000 times. The script was developed for Microsoft SQL Server 2012
or later. Data are stored in temp tables--a separate database does
need to be created to run the script.

This work was conducted with support from:

1) Harvard Catalyst | The Harvard Clinical and Translational 
Science Center (National Center for Research Resources and the 
National Center for Advancing Translational Sciences, National 
Institutes of Health Award UL1 TR001102) and financial contributions 
from Harvard University and its affiliated academic healthcare 
centers. The content is solely the responsibility of the authors 
and does not necessarily represent the official views of Harvard 
Catalyst, Harvard University and its affiliated academic healthcare 
centers, or the National Institutes of Health.

2) Informatics for Integrating Biology & the Bedside, which is 
sponsored by the National Institutes of Health Office of the 
Director, National Library of Medicine and the National Institute 
of General Medical Sciences (2U54LM008748).

3) Scalable Collaborative Infrastructure for a Learning Healthcare
System (SCILHS), which is funded by the Patient Centered Outcomes 
Research Institute (CDRN130604608).

This script is published under the GNU GPL v3 license:
http://www.gnu.org/licenses/gpl.txt

*/


--***********************************************************************
--***********************************************************************
-- RUN THE SIMULATION
--***********************************************************************
--***********************************************************************

-------------------------------------------------------------------------
-- Create a temp table with numbers 0 to 99,999
-------------------------------------------------------------------------

create table #n (
	n int primary key
)

;with n as (
	select 0 n union all select 1 union all select 2 union all select 3 union all select 4 union all select 5
	 union all select 6 union all select 7 union all select 8 union all select 9
)
insert into #n
	select a.n+10*b.n+100*c.n+1000*d.n+10000*e.n
	from n a, n b, n c, n d, n e

-------------------------------------------------------------------------
-- Create a temp table to store info about each hospital for each trial
-------------------------------------------------------------------------

create table #Hospitals (
	Trial int not null,
	Hospital int not null,
	RandNorm1 float,
	RandNorm2 float,
	RandNorm3 float,
	RandUniform float,
	Patients float,
	StartYear float,
	EndYear float,
	Downtime float
)
alter table #Hospitals add primary key (Trial,Hospital)

-------------------------------------------------------------------------
-- Generate 3 normally distributed random numbers for each hospital
-------------------------------------------------------------------------

insert into #Hospitals (Trial, Hospital, RandNorm1, RandNorm2, RandNorm3, RandUniform)
	select t.n, h.n, 
		sqrt(-2*log(abs(binary_checksum(newid())/2147483648.0)))
			*cos(2*pi()*abs(binary_checksum(newid())/2147483648.0)) RandNorm1,
		sqrt(-2*log(abs(binary_checksum(newid())/2147483648.0)))
			*cos(2*pi()*abs(binary_checksum(newid())/2147483648.0)) RandNorm2,
		sqrt(-2*log(abs(binary_checksum(newid())/2147483648.0)))
			*cos(2*pi()*abs(binary_checksum(newid())/2147483648.0)) RandNorm3,
		abs(binary_checksum(newid())/2147483648.0) RandUniform
	from #n t, #n h
	where t.n between 1 and 1000 
		and h.n between 1 and 4000

create unique nonclustered index idx_HopitalTrail on #Hospitals(Hospital,Trial)

-------------------------------------------------------------------------
-- Calculate the attributes of each hospital using the random numbers
-------------------------------------------------------------------------

update #Hospitals
	set Patients = log(Hospital/4001.0)/log(1/4001.0),
		StartYear = -exp(2.358109+0.340544*RandNorm1),
		EndYear = -exp(0.274653+0.788695*RandNorm2),
		Downtime = exp(-4.084964+1.120003*RandNorm3)

update #Hospitals
	set EndYear = 0
	where EndYear > 0

update #Hospitals
	set StartYear = EndYear
	where StartYear > EndYear

update #Hospitals
	set Downtime = (case when Downtime < 0 then 0 else 1 end)
	where Downtime < 0 or Downtime > 1

select Trial, Hospital, 0.5 * (1 - log(k)/log(4000)) Patients
	into #HospitalPatients
	from (
		select Trial, Hospital, row_number() over (partition by Trial order by newid()) k
		from #Hospitals
	) t

alter table #HospitalPatients add primary key (Trial,Hospital)

update h
	set h.Patients = p.Patients
	from #Hospitals h, #HospitalPatients p
	where h.Trial = p.Trial and h.Hospital = p.Hospital


--***********************************************************************
--***********************************************************************
-- GET THE RESULTS
--***********************************************************************
--***********************************************************************

-------------------------------------------------------------------------
-- Availability (By Number of Hospitals in Network)
-------------------------------------------------------------------------

select h NetworkSize, (1000000 - sum(n) over (order by h))/1000000.0 AllSitesAvailable
	from (
		select h, count(*) n
			from (
				select a, b, min(h) h
				from (
					select h.Trial a, r.Trial b, h.Hospital h, (case when r.RandUniform < h.Downtime then 0 else 1 end) Available
					from #Hospitals h, #Hospitals r
					where h.Hospital = r.Hospital
				) t
				where Available = 0
				group by a, b
			) t
			group by h
	) t
	order by h

-------------------------------------------------------------------------
-- Availability (4000 Hospital Network)
-------------------------------------------------------------------------

select x HospitalsAvailable, count(*)/1000000.0 Frequency
	from (
		select h.Trial a, r.Trial b, sum(case when r.RandUniform < h.Downtime then 0 else 1 end) x
		from #Hospitals h, #Hospitals r
		where h.Hospital = r.Hospital
		group by h.Trial, r.Trial
	) t
	group by x
	order by x

-------------------------------------------------------------------------
-- StartDate and EndDate distributions
-------------------------------------------------------------------------

declare @a table (m int, n int)
declare @b table (m int, n int)

insert into @a(m,n)
	select m, count(*) n
	from (
		select floor(StartYear*12) m
		from #Hospitals
	) t
	group by m

insert into @b(m,n)
	select m, count(*) n
	from (
		select floor(EndYear*12) m
		from #Hospitals
	) t
	group by m

select -(n.n+0.5)/12.0 [Year],
		isnull(a.n/maxA,0) [Start], 
		isnull(b.n/maxB,0) [End]
	from #n n
		left outer join @a a on a.m = -n.n
		left outer join @b b on b.m = -n.n
		cross join (select max(n)*1.0 maxA from @a) am
		cross join (select max(n)*1.0 maxB from @b) bm
	where n.n between 0 and 20*12
	order by n.n

-------------------------------------------------------------------------
-- Number of Patients with a Diagnosis by Year
-------------------------------------------------------------------------

select n [Year],
		avg(a*1.0) PatientsUniformModel,
		avg(b*1.0) PatientsSawtoothModel
	from (
		select h.Trial, -n.n n,
			sum(case when -n.n >= StartYear and -n.n < EndYear
					then Patients 
					else 0 end) a,
			sum(case when -n.n >= StartYear and -n.n < EndYear
					then Patients*(0.5 + 0.5*(-n.n-StartYear)/(EndYear-StartYear)) 
					else 0 end) b
		from #Hospitals h cross join #n n
		where n.n between 0 and 20
		group by h.Trial, n.n
	) t
	group by n
	order by n

-------------------------------------------------------------------------
-- Reproducability
-------------------------------------------------------------------------

-- Get the first day each hospital is not available

select a Trial, Hospital, 
		min(case when Available = 0 then b else 1001 end) u
	into #FirstUnavailable
	from (
		select h.Trial a, r.Trial b, h.Hospital, 
			(case when r.RandUniform < h.Downtime then 0 else 1 end) Available
		from #Hospitals h, #Hospitals r
		where h.Hospital = r.Hospital
	) t
	group by a, Hospital

alter table #FirstUnavailable add primary key (Trial,Hospital)

-- Add the first day that each hospital's data is updated

select h.Trial, h.Hospital, u.u, floor(-2*365*h.EndYear*h.RandUniform)+2 v
	into #Reproducibility
	from #Hospitals h, #FirstUnavailable u
	where h.Trial = u.Trial and h.Hospital = u.Hospital

alter table #Reproducibility add primary key (Trial,Hospital)

-- Determine the number of days a query is reproducible

;with r as (
	select Trial, Hospital, u, v, (case when u<v then u else v end) w,
		row_number() over (partition by Trial order by u) x,
		row_number() over (partition by Trial order by v) y,
		row_number() over (partition by Trial order by (case when u<v then u else v end)) z
	from #Reproducibility r
), u as (
	select Trial, u, 4000-max(x) x from r group by Trial, u
), v as (
	select Trial, v, 4000-max(y) y from r group by Trial, v
), w as (
	select Trial, w, 4000-max(z) z from r group by Trial, w
)
select d [Day], avg(u*1.0) SameSites, avg(v*1.0) SameData, avg(w*1.0) SameResult
	from (
		select a.n t, b.n d, 
			isnull(min(u.x) over (partition by a.n order by b.n),4000) u,
			isnull(min(v.y) over (partition by a.n order by b.n),4000) v,
			isnull(min(w.z) over (partition by a.n order by b.n),4000) w
		from #n a
			cross join #n b
			left outer join u on u.Trial = a.n and u.u = b.n
			left outer join v on v.Trial = a.n and v.v = b.n
			left outer join w on w.Trial = a.n and w.w = b.n
		where a.n between 1 and 1000
			and b.n between 1 and 1000
	) t
	group by d
	order by d
