/*

Federated Query Simulation
Copyright (C) 2012 Griffin M Weber

Griffin M Weber, MD, PhD
Beth Israel Deaconess Medical Center, Boston, MA
Harvard Medical School, Boston, MA

October 20, 2012

This script simulates a federated query to determine the number of
distinct patients in a network of hospitals that match certain criteria.
It repeats the simulation multiple times varying the different
configuration parameters of the simulation. For more information, please
see the manuscript "Federated queries of clinical data repositories: 
The sum of the parts doesn't equal the whole".

This script should be run in Microsoft SQL Server 2005 or later.
It requires about 1.5 GB storage and takes several hours to complete.

This script is published under the GNU GPL v3 license:
http://www.gnu.org/licenses/gpl.txt

*/


-- *********************************************************
-- * Create a table of sequential integers
-- *********************************************************

create table n (n int primary key)

;with a as (
	select 0 n
	union all select 1
	union all select 2 
	union all select 3 
	union all select 4 
	union all select 5 
	union all select 6 
	union all select 7 
	union all select 8 
	union all select 9 
)
insert into n(n)
	select a.n+b.n*10+c.n*100+d.n*1000+e.n*10000+f.n*100000+g.n*1000000+h.n*10000000
	from a, a b, a c, a d, a e, a f, a g, a h

-- *********************************************************
-- * Create a table to store the results
-- *********************************************************

create table FedAggResults (
	ExperimentID int identity(0,1) primary key,
	ExperimentDate datetime,
	NumberOfHospitals int,
	PatientOverlap float,
	ConceptPrevalence float,
	DataFactOverlap float,
	SampleSize float,
	AdaptiveCountingBucketBits float,
	UpperBoundNoPartitions float,
	UpperBoundWithPartitions float,
	EstimatedCount float,
	LowerBoundWithPartitions float,
	LowerBoundNoPartitions float,
	PatientExchanges float,
	DistinctPatientExchanges float,
	UpperBoundTotalPatients float,
	ActualNumberOfPatients float,
	LowerBoundTotalPatients float,
	AdaptiveCountingEstimate float
)

-- *********************************************************
-- * Declare variables
-- *********************************************************

-- Declare the simulation parameters.
declare @NumberOfHospitals int
declare @PatientOverlap float
declare @ConceptPrevalence float
declare @DataFactOverlap float
declare @SampleSize int
declare @AdaptiveCountingBucketBits int

-- Declare loop variables.
declare @RepeatLoop int
declare @ParameterLoop int
declare @IncrementLoop int

-- *********************************************************
-- * Create loops to run multiple simulations.
-- *********************************************************

-- Repeat everything multiple times.
select @RepeatLoop = 0
while @RepeatLoop < 10
begin

	-- Increment loop variable.
	select @RepeatLoop = @RepeatLoop + 1
	
	-- Loop through each parameter.
	select @ParameterLoop = 0
	while @ParameterLoop < 7
	begin
	
		-- Increment loop variable.
		select @ParameterLoop = @ParameterLoop + 1
	
		-- Increment the value of one parameter.
		select @IncrementLoop = 0
		while @IncrementLoop < 10
		begin
		
			-- Increment loop variable.
			select @IncrementLoop = @IncrementLoop + 1

-- *********************************************************
-- * Set the simulation parameters.
-- *********************************************************

-- Set the default parameter values.
select	@NumberOfHospitals = 5,				-- int >= 1
		@PatientOverlap = 0.2,				-- float 0-1
		@ConceptPrevalence = 0.01,			-- float 0-1
		@DataFactOverlap = 0.1,				-- float 0-1
		@SampleSize = 10,					-- int >= 1
		@AdaptiveCountingBucketBits = 10	-- int 1-30

-- Update one parameter based on the @ParameterLoop and @IncrementLoop values.
if @ParameterLoop = 1
	select @NumberOfHospitals = @IncrementLoop
if @ParameterLoop = 2
	select @PatientOverlap = @IncrementLoop/10.0
if @ParameterLoop = 3
	select @PatientOverlap = 0, @IncrementLoop = 10
if @ParameterLoop = 4
	select @ConceptPrevalence =
		(case @IncrementLoop 
			when 1 then 0.00001 
			when 2 then 0.0001 
			when 3 then 0.001 
			when 4 then 0.01 
			when 5 then 0.1 
			when 6 then 0.2 
			when 7 then 0.3 
			when 8 then 0.5 
			when 9 then 0.75 
			when 10 then 1.0 
			end)
if @ParameterLoop = 5
	select @DataFactOverlap = @IncrementLoop/10.0
if @ParameterLoop = 6
	select @DataFactOverlap = 0, @IncrementLoop = 10
if @ParameterLoop = 7
	select @SampleSize = 
		(case @IncrementLoop 
			when 1 then 1 
			when 2 then 2 
			when 3 then 3 
			when 4 then 5 
			when 5 then 10 
			when 6 then 20 
			when 7 then 50 
			when 8 then 100 
			when 9 then 500 
			when 10 then 1000000 
			end)

-- *********************************************************
-- * Assign patients to hospitals.
-- *********************************************************

-- Create a list of 1000000 patients evenly distributed across the hospitals with no overlap.
-- This represents the "primary" hospital of the patients.
select n Patient, n % @NumberOfHospitals Hospital
	into #PatientHospital
	from n
	where n < 1000000

-- Calculate how many overlaps we need to add.
declare @NumberOfOverlaps int
select @NumberOfOverlaps = floor((@NumberOfHospitals-1)*1000000*@PatientOverlap)

-- Randomly select overlap pairs (Patient, HospitalIncrement).
-- This represents the "secondary" hospitals where patients receive care.
select top(@NumberOfOverlaps) *
	into #SecondaryHospital
	from (
		select p.n Patient, h.n+1 HospitalIncrement
		from n p, n h
		where p.n < 1000000 and h.n < (@NumberOfHospitals-1)
	) t
	order by newid()

-- Add indexes to improve performance.
create nonclustered index idx_p on #PatientHospital(Patient)
create nonclustered index idx_p on #SecondaryHospital(Patient)

-- Add the secondary hospitals to the list of primary hospitals.
-- The result is the final list of patient-hospital assignments.
insert into #PatientHospital (Patient, Hospital)
	select s.Patient,
			(case	when p.Hospital + s.HospitalIncrement < @NumberOfHospitals 
					then p.Hospital + s.HospitalIncrement
					else p.Hospital + s.HospitalIncrement - @NumberOfHospitals 
					end) Hospital
		from #PatientHospital p, #SecondaryHospital s
		where p.Patient = s.Patient

-- *********************************************************
-- * Assign clinical concepts (facts) to patients.
-- *********************************************************

-- Randomly select patients based on the concept prevalence.
-- For each patient, randomly select one of her hospitals to assign the fact.
-- For each patient's other hospitals, also assign the fact if a random number is below the data fact overlap.
select Patient, Hospital
	into #PatientHospitalFactTemp
	from (
		select Patient, Hospital, 
			row_number() over (partition by Patient order by newid()) RandomHospitalSort,
			abs(binary_checksum(newid())/power(2.0,31)) RandomNumber
		from #PatientHospital
		where Patient in (
			-- Randomly select patients who have the fact.
			select top(cast(floor(1000000*@ConceptPrevalence) as int)) n 
			from n 
			where n < 1000000 
			order by newid()
		)
	) t
	where RandomHospitalSort = 1 or RandomNumber < @DataFactOverlap

-- Add an index to improve performance.
create clustered index idx_p on #PatientHospitalFactTemp(Patient)

-- Create a list of all patient hospital pairs that also indicates whether the fact has been assigned.
select a.Patient, a.Hospital,
		(case when b.Patient is null then 0 else 1 end) HasFact
	into #PatientHospitalFact
	from #PatientHospital a
		left outer join #PatientHospitalFactTemp b 
			on a.Patient=b.Patient and a.Hospital=b.Hospital

-- Add an index to improve performance.
create unique clustered index idx_ph on #PatientHospitalFact(Patient, Hospital)

-- *********************************************************
-- * Assign patients to partitions.
-- *********************************************************

-- Create a summary table with each patient-hospital pair and the HospitalPartition and FactPartition.
-- The HospitalPartition is the combination of hospitals to which a patient is assigned.
-- The FactPartition is the combination of hospitals to which a patient's facts are assigned.
-- Note that partition numbers range from 1 to 2^NumberOfHospitals-1.
-- Assume patients can be identified in multiple hospitals with 100% accuracy.
;with h as (
	-- Determine the HospitalPartition
	select Patient, sum(power(2,Hospital)) c, count(*) n
	from #PatientHospitalFact
	group by Patient
), f as (
	-- Determine the FactPartition
	select Patient, sum(power(2,Hospital)) c, count(*) n
	from #PatientHospitalFact
	where HasFact = 1
	group by Patient
)
select p.Patient, p.Hospital, p.HasFact,
		h.c HospitalPartition, h.n HospitalPartitionSize,
		IsNull(f.c,0) FactPartition, IsNull(f.n,0) FactPartitionSize
	into #Patient
	from #PatientHospitalFact p
		inner join h on p.Patient = h.Patient
		left outer join f on p.Patient = f.Patient

-- Add indexes to improve performance.
create unique clustered index idx_ph on #Patient(Patient, Hospital)
create unique nonclustered index idx_hcp on #Patient(Hospital, HospitalPartition, Patient)

-- *********************************************************
-- * Determine bounds without using the partitions.
-- *********************************************************

-- Combine the hospitals to get bounds on the number of patients and number of patients with facts.
select	max(NumPatients) LowerBoundPatients,
		sum(NumPatients) UpperBoundPatients, 
		max(NumPatientsWithFacts) LowerBoundPatientsWithFacts,
		sum(NumPatientsWithFacts) UpperBoundPatientsWithFacts
	into #BoundsNoPartitions
	from (
		-- For each hospital, calculate the number of patients and number of patients with facts.
		select Hospital, count(*) NumPatients, sum(HasFact) NumPatientsWithFacts
		from #Patient
		group by Hospital
	) t

-- *********************************************************
-- * Determine the bounds with partitions.
-- *********************************************************

-- For each HospitalParition in each Hospital, calculate the number of patients and number of patients with facts.
select Hospital, HospitalPartition, count(*) NumPatients, sum(HasFact) NumPatientsWithFacts
	into #HospitalHospitalPartition
	from #Patient
	group by Hospital, HospitalPartition
create unique clustered index idx_x on #HospitalHospitalPartition(Hospital, HospitalPartition)
create unique nonclustered index idx_y on #HospitalHospitalPartition(HospitalPartition, Hospital)

-- For each HospitalPartition, combine the hospitals to get the actual number of patients 
-- and bounds on the number of patients with facts.
select	HospitalPartition,
		max(NumPatients) NumPatients,
		max(NumPatientsWithFacts) LowerBoundPatientsWithFacts,
		(case when sum(NumPatientsWithFacts) < max(NumPatients)
			then sum(NumPatientsWithFacts)
			else max(NumPatients)
			end) UpperBoundPatientsWithFacts
	into #HospitalPartition
	from #HospitalHospitalPartition
	group by HospitalPartition
create unique clustered index idx_x on #HospitalPartition(HospitalPartition)

-- Combine the HospitalPartitions to get overall actual number of patients
-- and the overall bounds on the number of patients with facts.
select	sum(NumPatients) NumPatients,
		sum(LowerBoundPatientsWithFacts) LowerBoundPatientsWithFacts,
		sum(UpperBoundPatientsWithFacts) UpperBoundPatientsWithFacts
	into #BoundsWithPartitions
	from #HospitalPartition

-- *********************************************************
-- * Estimate the number of patients using sampling.
-- *********************************************************

-- Select patients in each HospitalPartition in each hospital up to the sample size.
select Hospital, HospitalPartition, HospitalPartitionSize, FactPartition, FactPartitionSize, Patient
	into #PatientSample
	from (
		-- Randomly order patients with facts.
		select Patient, Hospital, HospitalPartition, HospitalPartitionSize, FactPartition, FactPartitionSize,
			row_number() over (partition by Hospital, HospitalPartition order by newid()) k
		from #Patient
		where HasFact = 1
	) t
	where k <= @SampleSize

-- Add an index to improve performance.
create nonclustered index idx_x on #PatientSample(Hospital, HospitalPartition, FactPartition)

-- Calculate an estimated number of distinct patients with facts across all hospitals.
;with a as (
	-- For each FactPartition in each HospitalPartition in each hospital, calculate the number of sampled patients.
	select Hospital, HospitalPartition, FactPartition, FactPartitionSize, count(*) SampleCount
	from #PatientSample
	group by Hospital, HospitalPartition, FactPartition, FactPartitionSize
), b as (
	-- For each HospitalPartition in each hospital, calculate the number of sampled patients.
	select Hospital, HospitalPartition, count(*) SampleCount
	from #PatientSample
	group by Hospital, HospitalPartition
), c as (
	-- For each FactPartition in each HospitalPartition in each hospital, estimate the acutal number of patients with facts.
	-- Do this by dividing the FactPartition sample count by the fraction of patients with facts in the HospitalPartition that were sampled.
	select a.Hospital, a.HospitalPartition, a.FactPartition, a.FactPartitionSize,
		a.SampleCount / (b.SampleCount / (h.NumPatientsWithFacts * 1.0)) EstimatedPatientsWithFacts
	from a, b, #HospitalHospitalPartition h
	where a.Hospital=b.Hospital and a.HospitalPartition=b.HospitalPartition
		and a.Hospital=h.Hospital and a.HospitalPartition=h.HospitalPartition
), d as (
	-- For each FactPartition in each HospitalPartition, average the individual hospital estimates.
	select HospitalPartition, FactPartition,
		sum(EstimatedPatientsWithFacts)/(max(FactPartitionSize)*1.0) EstimatedPatientsWithFacts
	from c
	group by HospitalPartition, FactPartition
), e as (
	-- For each HospitalPartition, add the estimates from all the FactPartitions.
	select HospitalPartition, sum(EstimatedPatientsWithFacts) EstimatedPatientsWithFacts
	from d
	group by HospitalPartition
), f as (
	-- For each HospitalPartition, make sure the estimates are within the bounds.
	select e.HospitalPartition,
		(case	when e.EstimatedPatientsWithFacts < h.LowerBoundPatientsWithFacts 
					then h.LowerBoundPatientsWithFacts
				when e.EstimatedPatientsWithFacts < h.UpperBoundPatientsWithFacts
					then e.EstimatedPatientsWithFacts
				else h.UpperBoundPatientsWithFacts
				end) EstimatedPatientsWithFacts
	from e, #HospitalPartition h
	where e.HospitalPartition = h.HospitalPartition
)
-- Add the estimates from each HospitalPartition
select sum(EstimatedPatientsWithFacts) EstimatedPatientsWithFacts
	into #EstimatedPatientsWithFacts
	from f

-- *********************************************************
-- * Estimate the number of patients using adaptive counting.
-- *********************************************************

declare @Buckets int
select @Buckets = power(2,@AdaptiveCountingBucketBits)

;with Buckets as (
	select Bucket, max(BucketValue) BucketValue
	from (
		select Hospital, Bucket, max(BucketIndex) BucketValue
		from (
			select Hospital, Patient, Bucket, min(n) BucketIndex
			from (
				select *, PatientHash & (@Buckets-1) Bucket, PatientHash / @buckets BucketHash
				from (
					select *, abs(cast(HASHBYTES('SHA1', cast(Patient as varchar(50))) as int)) PatientHash
					from #Patient
					where HasFact > 0
				) t
			) t, n
			where (n < 31) and (BucketHash & power(2,n) > 0)
			group by Hospital, Patient, Bucket
		) t
		group by Hospital, Bucket
	) t
	group by Bucket
)
select (case when (select count(*) from Buckets) <= 0.949 * @Buckets
			then (select -@Buckets * log(1.00000-count(*)/(@Buckets*1.00000)) from Buckets)
			else (select power(2.000000,sum(BucketValue)/(@Buckets*1.000000)) * @Buckets * 0.79402 from Buckets)
			end) AdaptiveCountingEstimate
	into #AdaptiveCounting

-- *********************************************************
-- * Save results.
-- *********************************************************

insert into FedAggResults (
			ExperimentDate, 
			NumberOfHospitals, PatientOverlap, ConceptPrevalence, DataFactOverlap, SampleSize, AdaptiveCountingBucketBits,
			UpperBoundNoPartitions, UpperBoundWithPartitions, EstimatedCount, LowerBoundWithPartitions, LowerBoundNoPartitions, 
			PatientExchanges, DistinctPatientExchanges, 
			UpperBoundTotalPatients, ActualNumberOfPatients, LowerBoundTotalPatients,
			AdaptiveCountingEstimate
	)
	select	GetDate(),
			@NumberOfHospitals, @PatientOverlap, @ConceptPrevalence, @DataFactOverlap, @SampleSize, @AdaptiveCountingBucketBits,
			n.UpperBoundPatientsWithFacts,
			w.UpperBoundPatientsWithFacts,
			e.EstimatedPatientsWithFacts,
			w.LowerBoundPatientsWithFacts,
			n.LowerBoundPatientsWithFacts,
			(select sum(HospitalPartitionSize-1) from #PatientSample),
			(select count(*) from #PatientSample where HospitalPartitionSize > 1),
			n.UpperBoundPatients,
			w.NumPatients,
			n.LowerBoundPatients,
			a.AdaptiveCountingEstimate
		from #BoundsNoPartitions n, #BoundsWithPartitions w, #EstimatedPatientsWithFacts e, #AdaptiveCounting a

-- *********************************************************
-- * Drop temp tables created within the loop.
-- *********************************************************

drop table #PatientHospital
drop table #SecondaryHospital
drop table #PatientHospitalFactTemp
drop table #PatientHospitalFact
drop table #Patient
drop table #BoundsNoPartitions
drop table #HospitalHospitalPartition
drop table #HospitalPartition
drop table #BoundsWithPartitions
drop table #PatientSample
drop table #EstimatedPatientsWithFacts
drop table #AdaptiveCounting

-- *********************************************************
-- * Close loops.
-- *********************************************************

		end
	end
end

-- *********************************************************
-- * Display results.
-- *********************************************************

select ExperimentID % 52 ExpID, NumberOfHospitals, PatientOverlap, ConceptPrevalence, DataFactOverlap, SampleSize, 
		avg(UpperBoundNoPartitions) UpperBoundNoPartitions,
		avg(UpperBoundWithPartitions) UpperBoundWithPartitions,
		avg(EstimatedCount) EstimateUsingSampling,
		avg(LowerBoundWithPartitions) LowerBoundWithPartitions,
		avg(LowerBoundNoPartitions) LowerBoundNoPartitions, 
		avg(PatientExchanges) PatientCodeExchanges,
		stdev(EstimatedCount)/sqrt(count(*)*1.0) StandardErrorOfEstimate,
		avg(AdaptiveCountingEstimate) EstimateUsingAdaptiveCounting,
		stdev(AdaptiveCountingEstimate)/sqrt(count(*)*1.0) StandardErrorOfAdaptiveCounting,
		count(*) NumberOfRepeats
	from FedAggResults
	group by ExperimentID % 52, NumberOfHospitals, PatientOverlap, ConceptPrevalence, DataFactOverlap, SampleSize
	order by 1


